# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains Autotracer class
"""
__package__ = 'ssa.modules'

import logging
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta
from fnmatch import fnmatchcase
from typing import Iterator, Optional, NamedTuple, List
from urllib.parse import urlparse

import numpy as np

from .common import Common
from ..autotracing import disabled_users, is_edition_supported
from ..configuration import load_tunables
from ..configuration.schemes import autotracing_tunables_schema
from ..internal.utils import (
    is_xray_version_supported,
    sentry_init
)

try:
    import xray.shared_library as xray_lib
except ImportError:
    xray_lib = None


@dataclass
class URL:
    """URL data container"""
    uri: str  # full URL including domain and URI
    avg_duration: int  # average duration during the day
    max_duration: int  # maximum duration during the day
    total_slow_reqs: int  # total SLOW request count during the day
    total_reqs: int  # total request count during the day
    belongs_to: str  # domain
    belongs_to_user: str  # domain owner


class URLS(NamedTuple):
    """
    Representation of an URL
    """
    domain_name: str
    uri_path: str


def url_split(url: str) -> URLS:
    """
    Split URL into domain_name and uripath including query string
    :param url: URL of format protocol://domain/path;parameters?query#fragment
    :return: namedtuple URL(domain_name, uripath)
    """
    fragments = urlparse(url)
    qs = f'?{fragments.query}' if fragments.query else ''
    uri = f'{fragments.path}{qs}' if fragments.path else '/'
    _no_www_netloc = fragments.netloc.replace('www.', '')
    _no_port_netloc = _no_www_netloc.split(':')[0]
    return URLS(_no_port_netloc, uri)


class AutoTracer(Common):
    """
    SSA autotracing module implementation.
    """

    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger('auto_tracer')
        self.logger.info('AutoTracer enabled: %s', __package__)
        self.disabled_users = list()
        self.tasks_list = list()

    def __call__(self):
        if not is_edition_supported():
            self.logger.info('AutoTracer skipped: unsupported CL edition')
            return
        if not is_xray_version_supported():
            self.logger.info('AutoTracer skipped: unsupported X-Ray version')
            return
        if xray_lib is None:
            self.logger.info('AutoTracer skipped: unable to import X-Ray')
            return

        start_tool = xray_lib.start_autotracing
        if start_tool is not None:
            self.logger.info('AutoTracer started')
            # reload config
            super().__init__()
            self.logger.debug('AutoTracer loaded config: %s', self.config)
            for url in self.urls_scheduled():
                self.logger.info("Starting auto task for %s", url.uri)
                try:
                    start_tool(url=url.uri, tracing_count=self.request_number)
                except Exception as e:
                    # most likely XRayError
                    self.logger.error('Failed to start task: %s', str(e))

    @staticmethod
    def load_conf() -> dict:
        """
        Load configuration
        """
        return load_tunables('autotracing.json',
                             autotracing_tunables_schema)

    @staticmethod
    def gets_domaininfo(domain_name: str) -> Optional[object]:
        """
        Gets domain info for the specified domain
        """
        if xray_lib is None:
            return None
        return xray_lib.domain_info(domain_name)

    def nginx_is_enabled(self, user_name: str) -> bool:
        """
        Says if nginx is enabled for a specific user
        """
        if self.skip_nginx:
            if xray_lib is None:
                return False
            return xray_lib.NginxUserCache(user_name).is_enabled
        return False

    def per_server_limit_recalculated(self) -> int:
        """Recalculate limit per server taking into account running tasks"""
        running_count = len([task for task in self.tasks_list if
                             task['status'] == 'running' and task[
                                 'user'] == '*autotracing*'])
        self.logger.debug('Number of running autotracing tasks %s',
                          running_count)
        return self.per_server_limit - running_count

    def excludes_old_tasks(self, full_list: Optional[list] = None) -> list:
        """
        Excludes tasks older than N days from the general list of tasks
        """
        if full_list is None:
            full_list = xray_lib.tasks_list()
            self.logger.debug('Task list loaded %s', full_list)

        n_days_ago = datetime.now() - timedelta(
            days=self.min_retracing_inverval)
        return [task for task in full_list
                if task['createtime'] > int(n_days_ago.timestamp())]

    def exclude_thesame_urls(self, current_url: str) -> bool:
        """
        Excludes url from the list if it completely matches the current url
        or if domain names match and "*" follows the domain name in the list
        """
        # c - current parsed url
        c = url_split(current_url)
        for task_data in self.tasks_list:
            # t - task parsed url
            t = url_split(task_data['url'])
            direct_match = c.domain_name == t.domain_name and c.uri_path == t.uri_path
            wildcard_match = fnmatchcase(
                c.domain_name, t.domain_name) and fnmatchcase(c.uri_path,
                                                              t.uri_path)
            if direct_match or wildcard_match:
                self.logger.debug(
                    'Skipped: URL %s was traced recently. Matched by %s',
                    current_url, task_data['url'])
                return True
        return False

    def pass_by_density(self, url_total_reqs: list,
                        domain_total_reqs: list) -> bool:
        """Check that URL density passes given threshold"""
        if self.density:
            url_density = np.amin(
                np.corrcoef(url_total_reqs, domain_total_reqs))
            self.logger.debug('Calculated density %s', url_density)
            return url_density > self.density_threshold
        return True

    def pass_by_slowness_percentile(self, url_durations: list) -> bool:
        """
        The measure of "slowness" for URL is:
        at least N% of its requests take more than X seconds.
        N% -- self.slow_duration_percentage
        X -- self.slow_duration_threshold
        """
        reversed_percentile = 100 - self.slow_duration_percentage
        reversed_percentile_value = np.percentile(url_durations,
                                                  reversed_percentile)
        self.logger.debug('Calculated %sth percentile %s',
                          reversed_percentile,
                          reversed_percentile_value)
        return reversed_percentile_value >= self.slow_duration_threshold

    def pass_by_allowed_throttling(self, url_throttled_reqs: list,
                                   url_total_reqs: list) -> bool:
        """
        Check that percent of throttled requests per URL passes given threshold
        """
        throttled_percent = (sum(url_throttled_reqs)/sum(url_total_reqs))*100
        self.logger.debug('Calculated throttled percent %s', throttled_percent)
        return throttled_percent <= self.allowed_throttling_percentage

    def pass_by_engine(self, wp_status: bool) -> bool:
        """
        Check that URLs of a particular domain should be analyzed.
        For now we skip non-wordpress sites
        """
        if self.only_wp:
            return wp_status
        return True

    def urls_computationally_filtered(self,
                                      collected_stats: dict) -> Iterator[URL]:
        """
        Select all URLs suitable for auto tracing by very basic rules:
         - WP site
         - suitable throttling
         - suitable density
         - measure of "slow" URL
        """
        for domain, url_stats in collected_stats.items():
            domain_owner = None
            if not self.pass_by_engine(url_stats['is_a_wordpress_domain']):
                # skip non-wordpress sites
                self.logger.debug('Skipped by engine: non-wordpress')
                continue

            for url, data in url_stats.items():
                if url in self.non_url_fields:
                    # skip entry of domain requests counter
                    continue

                self.logger.debug('Processing URL %s', url)
                if not self.pass_by_allowed_throttling(
                        data['url_throttled_reqs'], data['url_total_reqs']):
                    self.logger.debug('Skipped by throttled percent')
                    continue

                if not self.pass_by_density(data['url_total_reqs'],
                                            url_stats['domain_total_reqs']):
                    self.logger.debug('Skipped by density')
                    continue

                if not self.pass_by_slowness_percentile(data['durations']):
                    self.logger.debug('Skipped by slowness percentile')
                    continue

                # the URL has passed all computational checks
                # from here we already need username
                if domain_owner is None:
                    domain_info = self.gets_domaininfo(domain)
                    if domain_info is None:
                        domain_owner = 'unknown'
                    else:
                        domain_owner = domain_info.user

                yield URL(url,
                          avg_duration=int(np.mean(data['durations'])),
                          total_slow_reqs=sum(data['url_slow_reqs']),
                          total_reqs=sum(data['url_total_reqs']),
                          max_duration=max(data['durations']),
                          belongs_to=domain,
                          belongs_to_user=domain_owner)

    def urls_selected(self,
                      stats_collected: Optional[dict] = None) -> Iterator[URL]:
        """
        From selected by computed thresholds URLs take those for which:
         - autotracing enabled
         - nginx disabled
        """
        if stats_collected is None:
            stats_collected, _ = self.load_stats()
            self.logger.debug('AutoTracer loaded data: %s', stats_collected)

        # fill disabled users list, initially empty
        self.fill_in_disabled_users()

        for url in self.urls_computationally_filtered(stats_collected):
            if url.belongs_to_user in self.disabled_users:
                # skip users for whom autotracing is disabled
                self.logger.debug('Skipped: autotracing is disabled for %s',
                                  url.belongs_to_user)
                continue
            if self.nginx_is_enabled(url.belongs_to_user):
                # skip url the owner of which uses nginx
                self.logger.debug('Skipped: nginx is enabled for %s',
                                  url.belongs_to_user)
                continue
            yield url

    def urls_scheduled(self) -> Iterator[URL]:
        """
        Schedule autotracing by sorted list taking into account the limits:
         - no same task for 10 days
         - limit per server
         - limit per domain
        Return resulting list of URLs scheduled for auto tracing
        """
        # initial limit counters
        general_tasks_counter = 0
        tasks_counter_per_domain = defaultdict(int)
        # sort selected URLs
        sorted_urls = self.urls_sorted()
        self.logger.debug('Sorted scheduled list %s', sorted_urls)
        # get task list and filter out all old tasks in one turn
        self.tasks_list = self.excludes_old_tasks()
        per_server_smart_limit = self.per_server_limit_recalculated()
        for url in sorted_urls:
            if self.exclude_thesame_urls(url.uri):
                # limit of days for url
                continue
            if general_tasks_counter < per_server_smart_limit:
                if tasks_counter_per_domain[url.belongs_to] < self.per_domain_limit:
                    general_tasks_counter += 1
                    tasks_counter_per_domain[url.belongs_to] += 1
                    yield url
                else:
                    self.logger.debug('Skipped URL %s by domain limit (%s)',
                                      url.uri,
                                      self.per_domain_limit)
            else:
                self.logger.debug('Skipped URL %s by server limit (%s)',
                                  url.uri,
                                  per_server_smart_limit)

    def urls_sorted(self) -> List[URL]:
        """
        Sort URLs by total number of requests first
        and by average duration second
        """
        first_serie = sorted(list(self.urls_selected()),
                             key=lambda u: u.avg_duration, reverse=True)
        return sorted(first_serie, key=lambda u: u.total_reqs,
                      reverse=True)

    def fill_in_disabled_users(self) -> None:
        """
        Fill internal list of disabled users
        """
        self.disabled_users = disabled_users()


if __name__ == "__main__":
    sentry_init()
    logging.basicConfig(filename='auto_tracer_standalone.log',
                        level=logging.DEBUG)
    t = AutoTracer()
    t()
